
import java.util.List;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.function.BiPredicate;
import java.util.function.IntBinaryOperator;

public class TestSimpleExample
{
   public static void main(String[] args)
   {
      final List<Triple> triples = Arrays.asList(
         new Triple(1, 2, 3),
         new Triple(1, 4, 4),
         new Triple(2, 2, 5),
         new Triple(2, 4, 6));

      triples.sort(new CmpTriples_A() ); // Use the ascending Comparator.
      System.out.println(triples);
      triples.sort(new CmpTriples_D() ); // Use the descending Comparator.
      System.out.println(triples);

      // Define "functionals" that are the < and > operators.
      final BiPredicate<Integer, Integer> lt = (x,y) -> x < y;
      final BiPredicate<Integer, Integer> gt = (x,y) -> x > y;

      // Sort the list of triples using each
      // of the functional operators.
      triples.sort( new CmpTriplesF(lt) ); // ascending
      System.out.println(triples);
      triples.sort( new CmpTriplesF(gt) ); // descending
      System.out.println(triples);

      // More compact notation.
      triples.sort( new CmpTriplesF((x,y) -> x < y) ); // ascending
      System.out.println(triples);
      triples.sort( new CmpTriplesF((x,y) -> x > y) ); // descending
      System.out.println(triples);

      // Sort descending using a product
      triples.sort( new CmpTriplesR( (x,y) -> x > y,
                                     (x,y) -> x * y ) );
   }
}


class CmpTriples_A implements Comparator<Triple>
{
   public int compare(Triple t1, Triple t2)
   {
      final int s1 = t1.first() + t1.second();
      final int s2 = t2.first() + t2.second();
      if (s1 < s2) return -1;
      if (s2 < s1) return  1;
      return 0;
   }
}

class CmpTriples_D implements Comparator<Triple>
{
   public int compare(Triple t1, Triple t2)
   {
      final int s1 = t1.first() + t1.second();
      final int s2 = t2.first() + t2.second();
      if (s1 > s2) return -1;
      if (s2 > s1) return  1;
      return 0;
   }
}


/**
   This comparator does the work of the above two camparators.

   This comparator takes a < or > operator as a "functional parameter".

   Unfortunately, Java's notation for functional parameters is awkward
   and not easy to read.
*/
class CmpTriplesF implements Comparator<Triple>
{
   // https://docs.oracle.com/en/java/javase/21/docs/api/java.base/java/util/function/BiPredicate.html
   private final BiPredicate<Integer, Integer> cmp;

   CmpTriplesF(BiPredicate<Integer, Integer> cmp)
   {
      this.cmp = cmp;
   }

   public int compare(Triple t1, Triple t2)
   {
      final int s1 = t1.first() + t1.second();
      final int s2 = t2.first() + t2.second();
      if (cmp.test(s1, s2)) return -1;
      if (cmp.test(s2, s1)) return  1;
      return 0;
   }
}


/**
   This comparator uses two functional parameters.
*/
class CmpTriplesR implements Comparator<Triple>
{
   private final BiPredicate<Integer, Integer> cmp; // Can be < or >.
   private final IntBinaryOperator reduce;          // Can be + or *.

   CmpTriplesR(BiPredicate<Integer, Integer> cmp, IntBinaryOperator reduce)
   {
      this.cmp = cmp;
      this.reduce = reduce;
   }

   public int compare(Triple t1, Triple t2)
   {
      final int r1 = reduce.applyAsInt(t1.first(), t1.second());
      final int r2 = reduce.applyAsInt(t2.first(), t2.second());
      if (cmp.test(r1, r2)) return -1;
      if (cmp.test(r2, r1)) return  1;
      return 0;
   }
}
